#ifndef REACTION_HPP
#define REACTION_HPP

#include <string>
#include <vector>

#include "Inhibiteur.hpp"


// définir les classe de la std qui seront utilisées
using std::string;
using std::vector;

class Reaction         
{
    public:
        string _ec;
        string _nomEnzyme;
        string _organisme;
        
        vector<string> _substrats;
        vector<string> _produits;
        
        vector<int> _stochiometrieSubstrats;
        vector<int> _stochiometrieProduits;
        
        bool _reversible;
        
        vector<float> _km;
        vector<float> _kcat;
        vector<float> _ph;
        
        vector<Inhibiteur> _inhibiteurs_NS;
        vector<Inhibiteur> _inhibiteurs_S;
        
        vector<Reaction*> _reactions_amonts;
        
        // Constructeurs
        Reaction(vector<string> const& infoReaction);
        
        Reaction(vector<string> const& infoReaction,
                 vector<Inhibiteur> const& vINS,
                 vector<Inhibiteur> const& vIS);
                 
        Reaction(string const& ec, string const& nomEnzyme,
                 string const& organisme,
                 vector<string> const& substrats, 
                 vector<string> const& produits,
                 vector<float> const& km, vector<float> const& kcat,
                 string const& reversible,
                 vector<Inhibiteur> const& inhibiteurs);
        
        // Méthodes
        void remplissageConstante(vector<float> & constantes, string valeurs, char const delimiteur);
        void remplissage_pH(string valeurs, char const delimiteur);
        void StockageMetabolite(vector<string> const& metabolites, string const& type);
        void StockageInhibiteur(vector<Inhibiteur> const& inhibiteurs);
        
        void ajoutInhibitions_S(Inhibiteur inhibiteur);
        void ajoutInhibitions_NS(Inhibiteur inhibiteur);

        bool memeReaction(Reaction const& autre);
        bool memeReaction(Reaction const* autre);
        bool memeInhibiteurs(Reaction* autre);
        
        
        // Opérateurs
        bool operator!=(Reaction const& autre);
        bool operator<(Reaction const& autre);      
};


#endif // REACTIONS_HPP


// KM
// La constante de Michaelis Km est la concentration en substrat pour laquelle
// la vitesse initiale de la réaction est à la moitié de la vitesse initiale
// maximale
// Une valeur élevée traduit une faible affinité de l'enzyme pour son substrat
// Une valeur faible montre au contraire une forte affinité
//
// Kcat
// La constante catalytique ou "turnover number" correspond au nombre de
// molécules substrats transformées en produit par seconde
