#ifndef RECHERCHE_HPP
#define RECHERCHE_HPP

#include <set>
#include <vector>
#include <string>

#include "Requete.hpp"
#include "Reaction.hpp"
#include "Inhibiteur.hpp"
#include "Metabolite.hpp"


// définir les classe de la std qui seront utilisées
using std::set;
using std::string;
using std::vector;

class Recherche
{
    public:
        Requete* _requete;
        
        // Stockage des reactions liée aux entrées
        vector<vector<Reaction> > _reactionsEntrees;
        // et aux sorties
        vector<vector<Reaction> > _reactionsSorties;
        
        vector<vector<Inhibiteur> > _inhibitionsEntrees;
        vector<vector<Inhibiteur> > _inhibitionsSorties;
        
        vector<Reaction*> _resultats;
        
        vector<Metabolite> _reactionsEntreesAmonts;
        vector<Metabolite> _reactionsSortiesAmonts;
        
        // Constructeur
        Recherche(Requete &requete);
        
        // Méthode
        
        //--/RECHERCHE
        void recherche();
        
        //------/Inhibitions
        void rechercheInhibitions(int niveau);
        
        void recherchePotentielsInhibiteurs(int niveau,
                set<string> & inhibiteursEntreesPotentiels);
                
        void rechercheInhibitions_2(int niveau, set<string> const& inhibiteursEntreesPotentiels);


        void rechercheInhibiteursSorties(int niveau);
        void rechercheEnzymesSortieAInhiber(int niveau, set<string>& enzymesSortieAInhiber);
        void rechercheInhibiteursSorties_2(int niveau, set<string> const& enzymesSortieAInhiber);
        
        //------/Réactions
        void rechercheReactions(int niveau);
        
        void rechercheSubstratsProduits( int niveau,
            set<string> & substratsPotentiels,
            set<string> & produitsPotentiels);
            
        void rechercheReactionsInhibees(int niveau,
            vector<string> const& infoReaction,
            vector<Reaction> & vecteurReactionsEntrees);

        void rechercheReactionsProduisantInhibiteur(int niveau,
            vector<string> const& infoReaction,
            vector<Reaction> & vecteurReactionsSorties);
  
        void rechercheReactionEntrees(vector<string> const& infoReaction,
            set<string> const& substratsPotentiels,
            vector<Reaction> & vecteurReactionsEntrees);
            
        void rechercheReactionSorties(vector<string> const& infoReaction,
            set<string> const& produitsPotentiels,
            vector<Reaction> & vecteurReactionsSorties);
            
        bool reactionValide(Reaction & reaction);
        bool inhibiteurValide(Inhibiteur & inhibiteur);
    
        set<Reaction*> rechercheAmont( 
                          vector<string> const& metabolitesCibles,
                          vector<Metabolite> const& vMetabolite);

        void affichage();
};

/*
Structure de _reactionsEntrees
    Les réactions vont de substrat à produit

    LES ENTREES   REACTION niveau 1        REACTION niveau 2

                                           reaction 5 ...
                                         /
                               produit 1 - reaction 6 ...
                             /
                  reaction 1 - produit 2 - reaction 7 ...
                /  
    entrée 1    - reaction 2 - produit 3 - reaction 8 ...
                \
                  reaction 3 - produit 4 - reaction 9 ...

    entrée 2    - reaction 4 - produit 5 - reaction 10 ...
                             \
                               produit 6 - reaction 11 ...
                                         \
                                           reaction 12 ...


Structure de _reactionsSorties
    Les réactions vont de produit à subtrat

    LES ENTREES   REACTION niveau 1        REACTION niveau 2

                                            reaction 5 ...
                                         /
                               substrat 1 - reaction 6 ...
                             /
                  reaction 1 - substrat 2 - reaction 7 ...
                /  
    sortie 1    - reaction 2 - substrat 3 - reaction 8 ...
                \
                  reaction 3 - substrat 4 - reaction 9 ...

    sortie 2    - reaction 4 - substrat 5 - reaction 10 ...
                             \
                               substrat 6 - reaction 11 ...
                                         \
                                            reaction 12 ...
*/




#endif