#ifndef REQUETE_HPP
#define REQUETE_HPP

#include <string>
#include <vector>
#include <sstream>

#include "Reaction.hpp"

// définir les classe de la std qui seront utilisées
using std::string;
using std::vector;
using std::stringstream;

// constante
enum CINETIQUE {KM, KCAT, PH, KI, IC50};
extern const char* NOMS_CINETIQUES[];
extern const int TAILLE_FENETRE_LECTURE;

class Requete
{
    public:
        vector<string> _entrees;
        vector<string> _sorties;        
        bool           _rechercheSortie;        
        int            _profondeur; // Nombre de réactions maximum séparants les
                                    // entrées des sorties
        bool           _rechercheProfondeursInferieures;
        string         _nom_banqueReactions; // la banque de réactions
        stringstream   _banqueReactions; // la banque de réactions
        bool           _bkms; // la banque de réaction est BKMS ?
        string         _nom_banqueInhibitions; // la banque d'inhibitions       
        stringstream   _banqueInhibitions; // la banque d'inhibitions       
        string         _fichierSortie; // Nom du reseau
        bool           _metatool; // format du réseau généré est en metatool ou pas :p
        string         _listeSortie; // liste des réactions du réseaux en csv
        bool           _rechercheInhibitions;
        static bool    _wrong_modifier_tag;
        
        // valeur des cinétiques réactions
        float          _kmMin;
        float          _kmMax;
        float          _kcatMin;
        float          _kcatMax;
        float          _phMin;
        float          _phMax;
        float          _kiMin;
        float          _kiMax;
        float          _ic50Min;
        float          _ic50Max;
        
        // attribut static de la class
        static bool    _verbose;
        static bool    _chrono;
        static bool    _stat;
        
        // Constructeur
        Requete() : 
            _rechercheSortie(false),
            _profondeur(2),
            _rechercheProfondeursInferieures(false),
            _bkms(false),
            _metatool(false),
            _rechercheInhibitions(false),
            _kmMin(CONSTANTE_MIN),
            _kmMax(CONSTANTE_MAX),
            _kcatMin(CONSTANTE_MIN),
            _kcatMax(CONSTANTE_MAX),
            _phMin(CONSTANTE_MIN),
            _phMax(CONSTANTE_MAX),
            _kiMin(CONSTANTE_MIN),
            _kiMax(CONSTANTE_MAX),
            _ic50Min(CONSTANTE_MIN),
            _ic50Max(CONSTANTE_MAX){
                _fichierSortie = "";
            };

        // Méthodes
        void recupererProfondeur(char* profondeur);
        void decompression(string type);
        void traitement_substrats_produits_BKMS(vector<string> const& metabolites,
                                                stringstream* banque);
        void intervalleConstante(CINETIQUE cinetique, string valeur);


        bool commandeValide();

        void afficher_aide(const char* skillfinder);

        void affichage_param();
};

#endif // REQUETEBRENDA_HPP