#include <iostream>

#include <zlib.h>

#include "Requete.hpp"
#include "fonctions.hpp"
#include "constantes.hpp"

// définir les classe de la std qui seront utilisées
using std::cerr;
using std::cout;
using std::endl;
using std::stoi; // de <string>

const char* NOMS_CINETIQUES[] = {"Km", "Kcat", "pH", "Ki", "IC50"};
const int TAILLE_FENETRE_LECTURE = 1000000;

// définition attribut static
bool Requete::_verbose = false;
bool Requete::_chrono = false;
bool Requete::_stat = false;
bool Requete::_wrong_modifier_tag = false;


/*
    récupére la profondeur demander par l'utilisateur et vérifier que c'est bien

    un entier compris entre 1 et 4
*/
void Requete::recupererProfondeur(char* profondeur)
{
    bool erreur = false;

    // la valeur de la profondeur doit être comprise entre 1 et 4, il ne peut donc
    // y avoir qu'un caractères dans profondeur
    if(profondeur[1] != '\0')
        erreur = true;

    // si le caractère est bien un entier entre 1 et 4 alors le sauvegarder
    if('1' <= profondeur[0] && profondeur[0] <= '4')
        _profondeur = profondeur[0] - '0';
    else
        erreur = true;


    if(erreur)   
    {
        cerr << "Invalid distance value: " << profondeur << "\n";
        exit(0);
    }
}


/*
    Decompresse un fichier et stocke l'information dans un stringstream
    
    fichierCompresse nom du fichier compressé lu
    type type de données du fichier reactions ou inhibitions
    
    decompresse stringstream contenant l'information stocké dans le fichier compressé
 */
void Requete::decompression(string type)
{
    stringstream* banque(0);

    const char* fichierCompresse;
    
    // définition de la variable de flux en fonction de type
    if(type == "reactions")
    {
        banque = &(_banqueReactions);
        fichierCompresse = _nom_banqueReactions.c_str();
    }
    else // type == "inhibitions"
    {
        banque = &(_banqueInhibitions);
        fichierCompresse = _nom_banqueInhibitions.c_str();
    }
    
    // ouverture du fichier compressé
    gzFile fichier = gzopen(fichierCompresse, "r");

    if(fichier == 0)
    {
        cerr << "/!\\ Unable to read the file " << fichierCompresse << endl;
        exit(0);
    }

    stringstream decompresse;
    bool lecture = true;                // continuer la lecture ?

    while(lecture)  // tant que l'on peut continuer la lecture
    {
        // initialisation d'une nouvelle fenetre
        unsigned char buffer[TAILLE_FENETRE_LECTURE];

        // Lecture des caracteres dans la fenetre et stockage dans buffer de ces derniers
        int nbBytesLus = gzread(fichier, buffer, TAILLE_FENETRE_LECTURE - 1);
        
        // Le derniers caracteres sera un "\0" pour fermer la chaine de caractere
        buffer[nbBytesLus] = '\0';

        // Ajout de buffer à decompresse
        decompresse << buffer;

        // Si fichierCompresse completement lu alors on arrete la lecture
        lecture = (nbBytesLus == TAILLE_FENETRE_LECTURE-1) || !(gzeof (fichier));
    }
    gzclose(fichier);
    string ligne;
    int c = 2;
    bool premiere_ligne = true;
    while(getline(decompresse, ligne))
    {
        if(ligne.size() > 0)
        {
            if(premiere_ligne && type == "reactions")
            {
                string colonnes_bkms = "ID EC_Number   Recommended_Name    Reaction    "
                                        "Reaction_ID_BRENDA Reaction_ID_KEGG    "
                                        "Reaction_ID_MetaCyc    Reaction_ID_SABIO_RK    "
                                        "BRENDA_Pathway_Name    KEGG_Pathway_ID KEGG_Pathway_Name   "
                                        "MetaCyc_Pathway_ID MetaCyc_Pathway_Name    "
                                        "Stoichiometry_Check    Missing_Substrate   "
                                        "Missing_Product    Commentary_KEGG Commentary_MetaCyc  Remark";
                
                // extraction et traitement de la premiere ligne
                string pl;
                if (ligne[ligne.size()-1] == '\r') // suppression de '\r'
                {
                    pl = ligne.substr(0,ligne.size()-1);
                }
                else
                    pl = ligne;
                
                // comparaison de la premier ligne avec celle présente dans le fichier de BKMS
                if(pl == colonnes_bkms)
                    _bkms = true;
            }
            if(_bkms && type == "reactions")
            {
                if(!premiere_ligne) // ne pas prendre en compte la premiere ligne
                {
                    vector<string> sligne = split(ligne, '\t');
                    
                    // code EC
                    *banque << sligne[1] << '\t';
                    
                    // Nom de l'enzyme
                    *banque << sligne[2] << '\t';
                    
                    // organisme = NA car pas décrit dans la base de donnée BKMS
                    *banque << "NA"      << '\t';
                    
                    // Reaction
                    vector<string> reaction = split(sligne[3],  " = ");
                    if(reaction.size() != 2)
                    {
                        reaction = split(sligne[3], "<=>");
                        if(reaction.size() != 2)
                        {
                            cerr << ligne << endl;
                            cerr << " Format error for line reaction " << c << endl;
                        }
                    }
                    vector<string> substrats = split(reaction[0], " + ");
                    vector<string> produits  = split(reaction[1], " + ");
                    
                    // les subtrats
                    traitement_substrats_produits_BKMS(substrats, banque);
                    
                    // les produits
                    traitement_substrats_produits_BKMS(produits, banque);
                    
                    // la reversibilité
                    *banque << "?" << '\t';
                    
                    // les km
                    *banque << "?";
                    for(size_t s = 1; s < substrats.size(); ++s)
                        *banque << ";?";
                    *banque << '\t';
                    
                    // les kcat
                    *banque << "?";
                    for(size_t s = 1; s < substrats.size(); ++s)
                        *banque << ";?";
                    *banque << '\t';
                    
                    //le pH
                    *banque << "?";
                    
                    // fin de la ligne
                    *banque << "\n";
                    ++c;
                }
            }
            else
            {
                if (ligne[ligne.size()-1] == '\r') // suppression de '\r'
                {
                    *banque << ligne.substr(0,ligne.size()-1) << '\n';
                }
                else
                {
                    *banque << ligne << '\n';
                }
            }
            
            if(premiere_ligne)
                premiere_ligne = false;
        }
    }
}

/*
    Mettre au bon format les substrats et produits des réactions provenant du tableau de 
    la base de donnée BKMS
*/
void Requete::traitement_substrats_produits_BKMS(vector<string> const& metabolites,
                                                 stringstream* banque)
{
    // ---- les substrats
    for(size_t m = 0; m < metabolites.size(); ++m)
    {
        size_t l = 0;
        bool lecture_stoechiometrie = true;
        bool stoechiometrie = false;
        string valeur_stoechiometrie = "";
        while(lecture_stoechiometrie)
        {
            if(estNumerique(metabolites[m][l]))
            {
                stoechiometrie = true;
                valeur_stoechiometrie += metabolites[m][l];
            }
            else
            {
                lecture_stoechiometrie = false;
                
                // si le caractère suivant n'est pas un espace alors la valeur
                // numérique précedemment stockée n'est pas une valeur de
                // stoechiometrie
                if(metabolites[m][l] != ' ')
                    stoechiometrie = false;
            }
            ++l;
        }
        
        // la stoechiometrie si elle est précisée
        if(stoechiometrie)
        {
            *banque << valeur_stoechiometrie << "|";
            
            // nom du metabolite après la valeur de stoechiometrie
            bool ecriture = false;
            for(size_t l = valeur_stoechiometrie.size(); l < metabolites[m].size(); ++l)
            {
                if(metabolites[m][l] != ' ')
                    ecriture = true;
                if(ecriture)
                    *banque << metabolites[m][l];
            }
        }
        else // pas de stoechiometrie à gérer
        {
            *banque << metabolites[m];
        }
        
        if(m < metabolites.size() - 1)
            *banque << "-+-";
        else
            *banque << '\t';
        
    }
}

void Requete::intervalleConstante(CINETIQUE cinetique, string valeur)
{
    bool erreur = false;

    // Choix des variables à modifier en fonction de la cinétique
    float* cstMin(0);
    float* cstMax(0);    
    switch(cinetique)
    {
        case KM :
            cstMin = &_kmMin; cstMax = &_kmMax; break;
        case KCAT :
            cstMin = &_kcatMin; cstMax = &_kcatMax; break;
        case PH :
            cstMin = &_phMin; cstMax = &_phMax; break;
        case KI :
            cstMin = &_kiMin; cstMax = &_kiMax; break;
        case IC50 :
            cstMin = &_ic50Min; cstMax = &_ic50Max; break;
    }

    // séparation de valeur en fonction de ':' séparant la plus petite et la plus grande valeur
    vector<string> vCst = split(valeur, ':');

    // cas ou seul la valeur minimum est donnée:
    // -param valMin
    // ou
    // -param valMin:
    if(vCst.size() == 1)
    {
        if(estFloat(vCst[0]))
            *cstMin = stof(vCst[0]);
        else
            erreur = true;
    }
    else if(vCst.size() == 2)
    {
        if(estFloat(vCst[0]))
            *cstMin = stof(vCst[0]);
        else if(vCst[0].size() > 0)
            erreur = true;
        //else
        // cas ou seul la valeur max est donnée:
        // -param :valMax
        // donc pas de changement à faire pour la valeur min

        // valeur maximale
        if(estFloat(vCst[1]))
            *cstMax = stof(vCst[1]);
        else
            erreur = true;
    }
    else
    {
        erreur = true;
    }

    if(erreur)
    {
        cerr << "/!\\ Invalid value for " << NOMS_CINETIQUES[cinetique] << " '" << valeur << endl;
        exit(0);
    }
}

/*
    Methode pour s'assurer que l'utilisateur à renseigné correctement des valeurs des 
    paramètres

    retourne true si tout va bien
*/
bool Requete::commandeValide()
{
    bool valide = true;

    // il faut qu'il y ait au moins une entrée
    if(_entrees.size() == 0)
    {
        cerr << "- Missing input metabolites\n";
        valide = false;
    }

    // si des sorties sont définies alors le noter
    _rechercheSortie = _sorties.size() > 0;


    // si aucune sortie n'est donnée alors la profondeur de recheche
    // ne peut pas dépasser 2
    if(!_rechercheSortie && _profondeur > 2)
    {
        cerr << "- The distance can't exceed 2 if there are no outputs to search for.\n";
        valide = false;
    }

    // tester l'accessibilité des banques de réactions d'inhibitions
    if(FILE* fichier = fopen(_nom_banqueReactions.c_str(),"r"))
        fclose(fichier);
    else
    {
        cerr << "- The enzymatic reaction file : " << _nom_banqueReactions << " is not accessible\n";
        valide = false;
    }
    
    if(_rechercheInhibitions)
    {
        if(FILE* fichier = fopen(_nom_banqueInhibitions.c_str(),"r"))
            fclose(fichier);
        else
        {
            cerr << "- The enzymatic inhibition file : " << _nom_banqueInhibitions << " is not accessible\n";
            valide = false;
        }
    }

    // vérifier que le fichier de sortie à bien été défini
    if(_fichierSortie == "")
    {
        cerr << "- The output file must be defined\n";
        valide = false;
    }

    // tester que les valeurs des cinétiques soient bien rentrées
    if(_kmMin > _kmMax)
    {
        cerr << "- the minimum Km value cannot be greater than the maximum value\n";
        valide = false;
    }
    
    if(_kcatMin > _kcatMax)
    {
        cerr << "- the minimum kcat value cannot be greater than the maximum value\n";
        valide = false;
    }

    if(_phMin > _phMax)
    {
        cerr << "- the minimum pH value cannot be greater than the maximum value\n";
        valide = false;
    }

    if(_kiMin > _kiMax)
    {
        cerr << "- the minimum pH value cannot be greater than the maximum value\n";
        valide = false;
    }

    if(_ic50Min > _ic50Max)
    {
        cerr << "- the minimum IC50 value cannot be greater than the maximum value\n";
        valide = false;
    }

    
    return valide;

}


void Requete::afficher_aide(const char* skillfinder)
{
    char message[4096];

    sprintf(message, 
"%s\n"
"\n"
"This program searches in a list of enzymatic reactions for those that link\n"
"input metabolites to output metabolites. It can also search for all\n"
"reactions consuming a list of input metabolites\n"
"\n"
"Usage:\n"
"======\n"
"- search reactions that connect input metabolites to output metabolites\n"
"%s -in <inputs> -out <outputs> -br <reactions file> -bi <inhibitions file> \\\n"
"   -o <output file> [OPTIONS]\n"
"\n"
"- search for reactions that consume input metabolites\n"
"%s -in <inputs> -br <reactions file> -bi <inhibitions file> \\\n"
"   -o <output file> [OPTIONS]\n"
"\n"
"Mandatory parameters:\n"
"=====================\n"
"  -in            input metabolites, separated by ';'\n"
"                 -in \"i1;i2\"\n"
"  -o             output file containing the reactions found, in SBML format\n"
"                 -o <output file>\n"
"  -br            file containing enzymatic reactions\n"
"                 -br <enzymatic reactions file>\n"
"  -bi            file containing enzymatic inhibitions\n"
"                 -br <enzymatic inhibitions file>\n"
"\n"
"Options:\n"
"========\n"
"  -out <file>    output metabolites, separated by ';'\n"
"                 -out \"o1;o2\"\n"
"  -d             distance, in number of reaction, between intput and output\n"
"                 metabolites. 1 <= distance <= 4\n"
"                   -default: 2\n"
"  -sd            Search for reactions also at distances smaller than d\n"
"  -km            filter on Km values by defining an interval (values in mM)\n"
"                 -km \"min:max\"\n"
"  -kcat          filter on kcat values by defining an interval\n"
"                 -km \"min:max\"\n"
"  -ph            filter on pH values by defining an interval\n"
"                 -ph \"min:max\"\n"
"  -ki            filter on Ki values by defining an interval (values in mM)\n"
"                 -ki \"min:max\"\n"
"  -ic50          filter on IC50 values by defining an interval (values in mM)\n"
"                 -ic50 \"min:max\"\n"
"  -l             generate a file in TSV format containing the list of\n"
"                 reactions found\n"
"                 -l <file>\n"
"  -ec            use EC numbers rather than enzyme names\n"
"\n"
"  -h, --help     print this help\n"
"  -v, --verbose  print informations during the execution\n"
, skillfinder, skillfinder, skillfinder);
    
    cout << message << endl;
    exit(0);
}

void Requete::affichage_param()
{
    cerr << "The inputs:\n\t";    
    for(auto& e : _entrees)
        cerr << e << "\t";
    cerr << "\n";

    cerr << "The outputs: \n\t";
    for(auto& s : _sorties)
        cerr << s << "\t";
    cerr << "\n";

    cerr << "Distance:\n\t" << _profondeur << "\n";

    cerr << "SBML output file:\t" << _fichierSortie << "\n";
    cerr << "TSV output file: \t" <<   _listeSortie << "\n";

    cerr << "\n";

    if(_rechercheInhibitions)
        cerr << "Search for inhibition\n";
    if(_rechercheProfondeursInferieures)
        cerr << "search at smaller distance\n";

    cerr << "\nFilters:\n";
    cerr << "KM   :\t" << _kmMin   << " " << _kmMax     << "\n";
    cerr << "Kcat :\t" << _kcatMin << " " << _kcatMax   << "\n";
    cerr << "pH   :\t" << _phMin   << " " << _phMax     << "\n";
    cerr << "Ki   :\t" << _kiMin   << " " << _kiMax     << "\n";
    cerr << "IC50 :\t" << _ic50Min << " " << _ic50Max   << "\n";
}
