#include <iostream>

#include "Requete.hpp" // Requete::_verbose
#include "Traitement.hpp"
#include "fonctions.hpp"


using std::cerr;
using std::endl;

Traitement::Traitement( Recherche &recherche)
{
    _recherche = &recherche;
    
    connexion();
    
    rechercheInhibition();
}


/*
    Selection des réactions qui permettent de connecter au moins une entrée à
    une sortie en fonction de la/les profondeur.s donnée.s
*/
void Traitement::connexion()
{
    switch(_recherche->_requete->_profondeur)
    {
        case 4:
            prodondeur4();
            if(!_recherche->_requete->_rechercheProfondeursInferieures)
                break;
        case 3:
            prodondeur3();
            if(!_recherche->_requete->_rechercheProfondeursInferieures)
                break;
        case 2:
            prodondeur2();
            if(!_recherche->_requete->_rechercheProfondeursInferieures)
                break;
        case 1:
            prodondeur1();
            if(!_recherche->_requete->_rechercheProfondeursInferieures)
                break;
    }
}

/*
    Selection des reactions qui consomme au moins une entrée et produitsent au moins une sortie

    Pour les réseaux de profondeur 1

    entrée -> réaction 1 -> sortie
*/
void Traitement::prodondeur1()
{
    int niveau = 0;    
    for(Reaction& reaction : _recherche->_reactionsEntrees[niveau])
    {
        if(elementCommun(_recherche->_requete->_sorties, reaction._produits))
        {
            insertReaction(_reactionsFinales, reaction);
        }
    }
}


/*
    Selection des reactions qui connectent au moins une entrée à une sortie en deux réactions

    Pour les réseaux de profondeur 2

    entree -> réaction 1 -> réaction 2 -> sortie
*/
void Traitement::prodondeur2()
{
    if(Requete::_verbose)
        cerr << "\nDistance 2:\n";
    connexionEntreNiveau(0, 0);
}   


/*
    De même mais avec trois réactions

    Pour les réseaux de profondeur 3

    entrée -> réaction 1 -> réaction 2 -> réaction 3 -> sortie
*/
void Traitement::prodondeur3()
{
    if(Requete::_verbose)
        cerr << "\nDistance 3:\n";
    connexionEntreNiveau(0, 1);
}

/*
    De même pour quatre réactions

    Pour les réseaux de profondeur 4

    entrée -> réaction 1 -> réaction 2 -> réaction 3 -> réaction 4 -> sortie
*/
void Traitement::prodondeur4()
{
    if(Requete::_verbose)
        cerr << "\nDistance 4:\n";
    connexionEntreNiveau(1, 1);
}



/*
    Fait la connexion entre les reactions d'entrées du niveau "niveauEntree" et
    et les réactions de sorties du niveau "niveauSortie"

    cf le schéma dans Recherche.hpp 
*/
int Traitement::connexionEntreNiveau(int niveauEntree, int niveauSortie)
{
    int nbInsertion = 0;
    
    // Identification de tous les produits des reactions du niveau donné liées aux entrées
    set<string> produits;
    for(Reaction& reactionE : _recherche->_reactionsEntrees[niveauEntree])
    {
        for( string produit : reactionE._produits)
        {
            // H2O est toujours en excé dans le milieu de réaction donc il ne faut pas le prendre en compte
            if(produit != "H2O") 
                produits.insert(produit);
        }
    }
    
    // Identification de tous les substrats des réactions du niveau donné liées aux sorties
    set<string> substrats;
    set<string> inhibiteursS;
    bool insertion = false;
    for(Reaction& reactionS : _recherche->_reactionsSorties[niveauSortie])
    {
        insertion = false;
        for( string substrat : reactionS._substrats)
        {
            // H2O est toujours en excé dans le milieu de réaction donc il ne faut pas le prendre en compte
            if(substrat != "H2O")
            {
                substrats.insert(substrat);
                // Si un substrat est présent parmis les produits identifiés précédemment
                // alors il y a connexion !
                if(produits.find(substrat) != produits.end())
                {
                    _metabolitesConnexion.insert(substrat);
                    
                    //Si la réaction qui fait connexion n'a pas encore été stocké
                    if(!insertion)
                    {
                        //le faire
                        insertReaction(_reactionSortiesFinales, reactionS);
                        
                        nbInsertion += insertReaction(_reactionsFinales, reactionS);

                        // et stocker tous les réactions de niveau inférieur liées à elle           
                        if(niveauSortie > 0)
                        {
                            nbInsertion += insertReaction(_reactionsFinales, reactionS._reactions_amonts);
                        }                       
                    }
                    insertion = true;
                }
            }
        }

        if(_recherche->_requete->_rechercheInhibitions)
        {
            for(Inhibiteur& inhibiteur : reactionS._inhibiteurs_NS)
            {
                if(inhibiteur._nom != "H2O")
                {
                    inhibiteursS.insert(inhibiteur._nom);
                    // si l'inhibiteur est aussi un des produits des reactions d'entrées alors il y a connexion
                    if(produits.find(inhibiteur._nom) != produits.end())
                    {
                        _metabolitesConnexion.insert(inhibiteur._nom);

                        //Si la réaction qui fait connexion n'a pas encore été stocké
                        if(!insertion)
                        {
                            //le faire
                            insertReaction(_reactionSortiesFinales, reactionS);
                            
                            nbInsertion += insertReaction(_reactionsFinales, reactionS);

                            // et stocker tous les réactions de niveau inférieur liées à elle           
                            if(niveauSortie > 0)
                            {
                                nbInsertion += insertReaction(_reactionsFinales, reactionS._reactions_amonts);
                            }                       
                        }
                        insertion = true;

                    }
                }
            }
            for(Inhibiteur& inhibiteur : reactionS._inhibiteurs_S)
            {
                if(inhibiteur._nom != "H2O")
                {
                    inhibiteursS.insert(inhibiteur._nom);
                    // si l'inhibiteur est aussi un des produits des reactions d'entrées alors il y a connexion
                    if(produits.find(inhibiteur._nom) != produits.end())
                    {
                        _metabolitesConnexion.insert(inhibiteur._nom);

                        //Si la réaction qui fait connexion n'a pas encore été stocké
                        if(!insertion)
                        {
                            //le faire
                            insertReaction(_reactionSortiesFinales, reactionS);
                            
                            nbInsertion += insertReaction(_reactionsFinales, reactionS);

                            // et stocker tous les réactions de niveau inférieur liées à elle           
                            if(niveauSortie > 0)
                            {
                                nbInsertion += insertReaction(_reactionsFinales, reactionS._reactions_amonts);
                            }                       
                        }
                        insertion = true;
                    }
                }
            }
        }
    }
    
    // Reparsage des produits de réactions d'entrées pour trouver les réactions d'entrées qui permettent
    // de faire la connexion
    for(Reaction& reactionE :_recherche->_reactionsEntrees[niveauEntree])
    {
        insertion = false;
        for( string produit : reactionE._produits)
        {
            // H2O est toujours en excé dans le milieu de réaction donc il ne faut pas le prendre en compte
            if( produit != "H2O" && substrats.find(produit) != substrats.end() && !insertion)
            {
                // Stockage de la réactions
                insertReaction(_reactionEntreesFinales, reactionE);
                
                nbInsertion += insertReaction(_reactionsFinales, reactionE);
                
                // stockage des réactions de niveau inférieur liées à elle
                if(niveauEntree > 0 )
                {
                    nbInsertion += insertReaction(_reactionsFinales, reactionE._reactions_amonts);
                }
                insertion = true;
            }
            if(_recherche->_requete->_rechercheInhibitions)
            {
                if(produit != "H2O" && inhibiteursS.find(produit) != inhibiteursS.end() && !insertion)
                {
                    // Stockage de la réactions
                    insertReaction(_reactionEntreesFinales, reactionE);
                    
                    nbInsertion += insertReaction(_reactionsFinales, reactionE);
                    
                    // stockage des réactions de niveau inférieur liées à elle
                    if(niveauEntree > 0 )
                    {
                        nbInsertion += insertReaction(_reactionsFinales, reactionE._reactions_amonts);
                    }
                    insertion = true;
                }
            }
        }
    }
    
    if(Requete::_verbose)
    {
        cerr << "Number of intermediate input products:    " << produits.size()     << endl;
        cerr << "Number of intermediate output substrates: " << substrats.size()    << endl;
        cerr << "Number of intermediate output inhibitors: " << inhibiteursS.size() << endl;
    }

    return nbInsertion;
}


/*
    Recherche des inhibitions entre les métabolites et les enzymes du réseau
    Les inhibiteurs seront rajoutés dans les réactions
    
    inhibiteursDataBase     chemin vers la table des inhibiteurs
*/
void Traitement::rechercheInhibition()
{
    // metabolites est une variable qui va stocker tous les produits et substrats des réactions
    set<string> metabolites;
    
    // ecs contient tous les code ec du réseaux
    set<string> ecs;
    
    // boucle sur toutes les réactions qui font la connexion
    for(size_t r = 0; r < _reactionsFinales.size(); r++)
    {
        // stockage du ec
        ecs.insert(_reactionsFinales[r]->_ec);
        
        // boucle sur les substrats de la r ieme réaction
        for(size_t s = 0; s < _reactionsFinales[r]->_substrats.size(); s++)
            metabolites.insert(_reactionsFinales[r]->_substrats[s]);
        
        // boucle sur les produits de la r ieme réaction
        for(size_t p = 0; p < _reactionsFinales[r]->_produits.size(); p++)
            metabolites.insert(_reactionsFinales[r]->_produits[p]);
    }
    
    // afficher le nom des enzymes du réseau
    // if(Requete::_verbose)
    //     for(set<string>::iterator it = ecs.begin(); it != ecs.end(); it++)
    //         cerr << *it << endl;
    
    // Lecture de la banque d'inhibiteurs
    string ligne;
    vector<string> splitLigne;
    
    set<string>::iterator it_ec = ecs.begin();
    
    /*
    La banque de d'inhibition est classé par ordre alphabétique ce qui facilite grandement la recherche des inhibteurs
    des reactions car le code EC des réactinos se trouve au début de la ligne
    */
    while(getline(_recherche->_requete->_banqueInhibitions, ligne) && it_ec != ecs.end())  // lecture du fichier
    {
        splitLigne = split(ligne, '\t');
        /* 
            splitLigne[0] = ec
            splitLigne[1] = organisme
            splitLigne[2] = inhibiteur
            splitLigne[3] = KI
            splitLigne[4] = IC50
        */
        
        if(splitLigne[0] == *it_ec)
        {
            if(metabolites.find(splitLigne[2]) != metabolites.end())
            {
                for(size_t r = 0; r < _reactionsFinales.size(); r++)
                {
                    if(_reactionsFinales[r]->_ec == splitLigne[0])
                    {   
                        Inhibiteur nouveauInhibiteur(splitLigne);
                        
                        if(_reactionsFinales[r]->_organisme == splitLigne[1])
                        {
                            _reactionsFinales[r]->_inhibiteurs_S.push_back(nouveauInhibiteur);
                        }
                        else
                        {
                            _reactionsFinales[r]->_inhibiteurs_NS.push_back(nouveauInhibiteur);
                        }
                    }
                }
            }
        }
        else // splitLigne[0] != *it_ec
        {
            it_ec++;
            if(splitLigne[0] == *it_ec)
            {
                if(metabolites.find(splitLigne[2]) != metabolites.end())
                {
                    for(size_t r = 0; r < _reactionsFinales.size(); r++)
                    {
                        if(_reactionsFinales[r]->_ec == splitLigne[0])
                        {
                            
                            Inhibiteur nouveauInhibiteur(splitLigne);
                            
                            if(_reactionsFinales[r]->_organisme == splitLigne[1])
                            {
                                _reactionsFinales[r]->_inhibiteurs_S.push_back(nouveauInhibiteur);
                            }
                            else
                            {
                                _reactionsFinales[r]->_inhibiteurs_NS.push_back(nouveauInhibiteur);
                            }
                        }
                    }
                }
            }
        }       
    }   
}
