#include <iostream>
#include <algorithm> // find()
#include <sstream>   // stringstream
#include <fstream>   // ofstream
// #include <stdio.h>

#include "Requete.hpp" // Requete::_verbose
#include "ecritureReseau.hpp"
#include "Reaction.hpp"
#include "fonctions.hpp"

// dfinir les classe de la std qui seront utilises
using std::cerr;
using std::endl;
using std::stringstream;
using std::ofstream;


#ifdef _WIN32
    char SEP = '\\';
    #include <direct.h>
    #define repertoireActuel _getcwd
#else
    char SEP = '/';
    #include <unistd.h>
    #define repertoireActuel getcwd
#endif

extern bool	use_ec_number;		// wether use EC number rather than enzyme name.

// --- Ecriture du rseau au format XML --- //

/*
    Ecriture du reseau trouv au format SBML/XML
    
    reponse     contient les ractions slectionnes
    fichierSortie   nom du fichier de sortie
*/
void ecritureReseau_xml(vector<Reaction*> & reponse, string const& fichierSortie)
{
    // Initialisation des variables de stockage
    string listeEspeces;
    string listeReactions;

    // Ecriture de l'entte du fichier SBML/XML
    string entete = ecritureEntete_xml(fichierSortie);

    // Ecriture de la liste des compartiments
    string compartiment = ecritureCompartiment_xml();

    // Lecture de reponse pour ecrire la liste des especes, la liste des reactions
    lecture_xml(reponse, listeEspeces, listeReactions);

    // Ecriture dans un fichier XML !
    ecritureReseau_xml(fichierSortie, entete, compartiment, listeEspeces, listeReactions);
}


/*
    Lecture du vecteur de reactions
    
    reponse      Contient les reactions slectionnes
    listeEspeces    stocke le texte qui contient la liste des "especes"/ractif pour le fichier SBML
    listeReactions  De mme pour les ractions
*/
void lecture_xml(vector<Reaction*> & reponse, string & listeEspeces, string & listeReactions)
{
    int compteurEspecesReactions = 0;

    map<string, string> especeToId;
    
    vector<Reaction*> reactionsEcrites;

    for(size_t i = 0; i < reponse.size(); i++)
    {       
        // Associer chaque especes de la raction  un identifiant
        remplirEspeceToId_xml(reponse[i], especeToId, compteurEspecesReactions);
    }
    // Ecriture de la liste des ractions
    listeReactions = ecritureReactions_xml(reponse, especeToId, reactionsEcrites, compteurEspecesReactions);

    // Ecrire la liste des lments
    listeEspeces = "        <listOfSpecies>\n";
    for (map<string, string>::iterator it  = especeToId.begin(); it != especeToId.end(); it++)
    {
        // name = it->first;
        // id   = it->second;
        listeEspeces += "            <species id=\"" + it->second + "\" name=\"" + it->first + "\" compartment=\"default\">\n"
                        "            </species>\n";
    }
    listeEspeces += "        </listOfSpecies>\n";
    
    if(Requete::_verbose)
        cerr << "Number of reactions found: " << reactionsEcrites.size() << endl;
}


/*
    Stocke au fur et mesure les noms des espces biologiques et leurs ID associs
    
    reaction    Contient les informations ncessaire pour dcrire une raction
    especeToId      Un map qui contient les nom des espces et leurs ID
    compteurEspeces Compte le nombre d'espce pour crer de nouveau ID
*/
void remplirEspeceToId_xml(Reaction* const& reaction, map<string,string> & especeToId, int & compteurEspeces)
{
    // Nom de l'enzyme
	if (use_ec_number)
	    remplirEspeceToId2_xml(reaction->_ec, especeToId, compteurEspeces);
	else
	    remplirEspeceToId2_xml(reaction->_nomEnzyme, especeToId, compteurEspeces);
    // Les substrats
    for(size_t i = 0; i < reaction->_substrats.size(); i++)
    {
        remplirEspeceToId2_xml(reaction->_substrats[i], especeToId, compteurEspeces);
    }
    // Les produits
    for(size_t i = 0; i < reaction->_produits.size(); i++)
    {
        remplirEspeceToId2_xml(reaction->_produits[i], especeToId, compteurEspeces);
    }
    // Les inhibiteurs non spcifiques
    for(size_t i = 0; i < reaction->_inhibiteurs_NS.size(); i++)
    {
        remplirEspeceToId2_xml(reaction->_inhibiteurs_NS[i]._nom, especeToId, compteurEspeces);
    }
    // Les inhibiteurs spcifiques
    for(size_t i = 0; i < reaction->_inhibiteurs_S.size(); i++)
    {
        remplirEspeceToId2_xml(reaction->_inhibiteurs_S[i]._nom, especeToId, compteurEspeces);
    }
}
/*
    Cr une nouvelle association Espece-ID dans especeToId

    cible       Nouvelle espce biologique
    especeToId  Un map qui contient les nom des espces et leurs ID
    compteurEspeces Compte le nombre d'espce pour crer de nouveau ID
*/
void remplirEspeceToId2_xml(string const& cible, map<string,string> & especeToId, int & compteurEspeces)
{
    if( !(especeToId.count(cible) > 0) )
    {
        compteurEspeces ++;
        stringstream ss;
        ss << "s" << compteurEspeces;
        string id = ss.str();
        especeToId[cible] = id;
    }
}


/*
    crit la liste des reactions
    
    reaction objet contenant les informations ncessaire pour dcrire la raction
    listeReactions string dcrivant au format XML les ractions
    especeToId pour faire la correspondance entre le nom de l'espece biologique et son ID
*/
string ecritureReactions_xml(vector<Reaction*> & reponse, map<string,string> & especeToId, vector<Reaction*>& reactionsEcrites, int & compteurReactions)
{

    string listeReactions = "        <listOfReactions>\n";

    vector<Reaction*>::iterator it_reponse_1 = reponse.begin();
    vector<Reaction*>::iterator it_reponse_2;

    while(it_reponse_1 != reponse.end())
    {
        // Silicell maker ne prend pas en charge le bon tag pour les modifieurs (vrai le 06-2022)
        string modifier_tag;
        if(Requete::_wrong_modifier_tag)
            modifier_tag = "speciesReference";
        else
            modifier_tag = "modifierSpeciesReference";


        reactionsEcrites.push_back((*it_reponse_1));

        it_reponse_2 = it_reponse_1 + 1;
        // La raction est elle rversible ?
        string reversible = ((*it_reponse_1)->_reversible) ? "true" : "false";
        listeReactions += "            <reaction id=\"re" + itos(++compteurReactions) + "\" reversible=\"" + reversible + "\">\n";

        // Ajout des substrats
        listeReactions += "                <listOfReactants>\n";
        for(size_t sub = 0; sub < (*it_reponse_1)->_substrats.size(); sub++)
        {
            string subId = especeToId[(*it_reponse_1)->_substrats[sub]];
            listeReactions += "                    <speciesReference species=\"" + subId + "\">\n";
            listeReactions += "                    </speciesReference>\n";
        }
        listeReactions += "                </listOfReactants>\n";

        // Ajout des produits
        listeReactions += "                <listOfProducts>\n";
        for(size_t prod = 0; prod < (*it_reponse_1)->_produits.size(); prod++)
        {
            string prodId = especeToId[(*it_reponse_1)->_produits[prod]];
            listeReactions += "                    <speciesReference species=\"" + prodId + "\">\n";
            listeReactions += "                    </speciesReference>\n";
        }
        listeReactions += "                </listOfProducts>\n";

        // Ajout des "modifieurs" - Enzyme
        listeReactions += "                <listOfModifiers>\n";
        string enzId = use_ec_number ? especeToId[(*it_reponse_1)->_ec] : especeToId[(*it_reponse_1)->_nomEnzyme];

        listeReactions += "                    <"+modifier_tag+" species=\"" + enzId + "\" sboTerm=\"SBO:0000460\" />\n";

        // Recherche d'autre raction identiques!
        vector<string> ecEcrits = {(*it_reponse_1)->_ec};
        while(it_reponse_2 != reponse.end())
        {
            if((*it_reponse_1)->memeReaction(*it_reponse_2) && (*it_reponse_1)->memeInhibiteurs((*it_reponse_2)))
            {
                if(find(ecEcrits.begin(), ecEcrits.end(), (*it_reponse_2)->_ec ) == ecEcrits.end())
                {
                    string id_enzyme = use_ec_number ? especeToId[(*it_reponse_2)->_ec] : especeToId[(*it_reponse_2)->_nomEnzyme];

                    listeReactions += "                    <"+modifier_tag+" species=\"" + id_enzyme + "\" sboTerm=\"SBO:0000460\" />\n";
                    ecEcrits.push_back((*it_reponse_2)->_ec);
                }
                it_reponse_2 = reponse.erase(it_reponse_2);
            }
            else if(it_reponse_2 != reponse.end())
                it_reponse_2++;
        }

        vector<string> inhibiteurDejaEcrit;
        
        // Ajout des "modifieurs" - Inhibiteurs spcifiques
        for(size_t inhi = 0; inhi < (*it_reponse_1)->_inhibiteurs_S.size(); inhi++)
        {
            string idInhi = especeToId[(*it_reponse_1)->_inhibiteurs_S[inhi]._nom];
            if(!estPresent(idInhi, inhibiteurDejaEcrit))
            {
                inhibiteurDejaEcrit.push_back(idInhi);
                listeReactions += "                    <"+modifier_tag+" species=\"" + idInhi + "\" sboTerm=\"SBO:0000020\" />\n";
            }
        }

        // Ajout des "modifieurs" - Inhibiteurs non spcifiques
        for(size_t inhi = 0; inhi < (*it_reponse_1)->_inhibiteurs_NS.size(); inhi++)
        {
            string idInhi = especeToId[(*it_reponse_1)->_inhibiteurs_NS[inhi]._nom];
            if(!estPresent(idInhi, inhibiteurDejaEcrit))
            {
                inhibiteurDejaEcrit.push_back(idInhi);
                listeReactions += "                    <"+modifier_tag+" species=\"" + idInhi + "\" sboTerm=\"SBO:0000020\" />\n";
            }
        }

        listeReactions += "                </listOfModifiers>\n";
        listeReactions += "            </reaction>\n";

        it_reponse_1++;
    }   

    listeReactions += "        </listOfReactions>\n";
    return listeReactions;
}

/*
    Ecriture de l'entere du fichier SBML

    fichierSortie   nom du fichier de sortie
*/
string ecritureEntete_xml(string const& fichierSortie)
{
    // string nomSortieAbsolu = nomAbsolu_xml(fichierSortie);

    string entete =
    "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
    "<sbml xmlns=\"http://www.sbml.org/sbml/level2/version4\" level=\"2\" version=\"4\">\n"
    "    <model>\n"
    "        <listOfUnitDefinitions>\n"
    "            <unitDefinition id=\"substance\" name=\"substance\">\n"
    "                <listOfUnits>\n"
    "                    <unit kind=\"mole\"/>\n"
    "                </listOfUnits>\n"
    "            </unitDefinition>\n"
    "            <unitDefinition id=\"volume\" name=\"volume\">\n"
    "                <listOfUnits>\n"
    "                    <unit kind=\"litre\"/>\n"
    "                </listOfUnits>\n"
    "            </unitDefinition>\n"
    "            <unitDefinition id=\"length\" name=\"length\">\n"
    "                <listOfUnits>\n"
    "                    <unit kind=\"metre\"/>\n"
    "                </listOfUnits>\n"
    "            </unitDefinition>\n"
    "            <unitDefinition id=\"time\" name=\"time\">\n"
    "                <listOfUnits>\n"
    "                    <unit kind=\"second\"/>\n"
    "                </listOfUnits>\n"
    "            </unitDefinition>\n"
    "        </listOfUnitDefinitions>\n";

    return(entete);
}


/*
    retourne le chemin absolu du fichier de sortie

    fichierSortie   nom du fichier de sortie
    cheminAbsolu    le chemin absolu du fichier de sortie
*/
string nomAbsolu_xml(string const& fichierSortie)
{
    // Initialise chemin a partir duquel le programme est execut
    // avec la fonction repertoireActuel() - cf haut du programme
    char cheminActuel[FILENAME_MAX];
    if (!repertoireActuel(cheminActuel, sizeof(cheminActuel)))
    {
        cerr << "/!\\ Unable to obtain absolute path of output file";
        exit(0);
    }
    string sCheminActuel(cheminActuel); // conversion en string

    // regarde si il n'y a pas de "retour au repertoire precedent dans fichierSortie"
    vector<string> splitSortie = split(fichierSortie, SEP);
    int compteurRetour = 0;
    for(size_t i = 0; i < splitSortie.size(); i++)
    {
        if(splitSortie[i] == "..")
        {
            compteurRetour++;
        }
    }

    // s'il n'y a pas de "retour" et fichierSortie est un chemin alors fichierSortie est deja un chemin absolu
    if(compteurRetour == 0 && splitSortie.size() > 1)
    {
        return(fichierSortie);
    }
    else
    {
        string cheminAbsolu; //string qui contiendra le chemin absolu

        // parcours de cheminActuel pour ecrire debut chemin absolu en ne prenant pas en compte les repertoires "inutiles"
        vector<string> splitCheminActuel = split(sCheminActuel, SEP);
        for(size_t i = 0; i < splitCheminActuel.size() - compteurRetour; i++)
        {
            cheminAbsolu += splitCheminActuel[i] + SEP;
        }

        // Ajout de fichierSortie
        for(size_t i = compteurRetour; i < splitSortie.size(); i++)
        {
            if(i < splitSortie.size() - 1) // pour ne pas rajouter de separateur a la fin
            {
                cheminAbsolu += splitSortie[i] + SEP;
            }
            else if(i == splitSortie.size() - 1)
            {
                cheminAbsolu += splitSortie[i];
            }
        }

        return(cheminAbsolu);
    }
}

/*
    Ecriture de la liste des compartimentspour le fichier SBML
*/
string ecritureCompartiment_xml()
{
    string listeCompartiment =
    "        <listOfCompartments>\n"
    "            <compartment id=\"default\" size=\"1\" units=\"volume\">\n"
    "            </compartment>\n"
    "        </listOfCompartments>\n";

    return(listeCompartiment);
}

/*
    Ecriture du reseau

    fichierSortie   nom du fichier de sortie
    entete          contenu de l'entete
    compartiment    description des compartiments
    listeEspeces    liste des espces biologiques et leur correspondance avec leur ID
    listeReactions  liste dcrivants les reactions
*/
void ecritureReseau_xml(string const& fichierSortie, string const& entete, string const& compartiment,
                        string const& listeEspeces,  string const& listeReactions)
{

    ofstream fluxSortie(fichierSortie);

    fluxSortie << entete << compartiment << listeEspeces << listeReactions << "    </model>\n</sbml>\n";

    fluxSortie.close();
}

// --- Ecriture du rseau au format MTO simplifi --- //
void ecritureReseau_mto(vector<Reaction*> const& reponse, string const& fichierSortie)
{
    // ouverture du flux pour l'ecriture
    ofstream fluxSortie(fichierSortie);


    if(fluxSortie)
    {
        // premier substrat ou produit ecrit
        bool premier;

        for(auto& reac : reponse)
        {
            // Code de l'enzyme
            fluxSortie << reac->_ec << " : ";

            // Substrats
            premier = true;
            for(auto& sub : reac->_substrats)
            {
                if(premier)
                {
                    fluxSortie << sub;
                    premier = false;
                }
                else
                    fluxSortie << " + " << sub;
            }

            // Reversibilit
            if(reac->_reversible)
                fluxSortie << " = ";
            else
                fluxSortie << " => ";

            // Produits
            premier = true;
            for(auto& prod : reac->_produits)
            {
                if(premier)
                {
                    fluxSortie << prod;
                    premier = false;
                }
                else
                    fluxSortie << " + " << prod;
            }
            fluxSortie << endl;
        }
    }
    else
    {
        cerr << "Unable to open " << fichierSortie << endl;
        exit(0);
    }
}
