#include <iostream>
#include <cstring>   // strcmp() 
#include <algorithm> // sort()
#include <sstream>

#include "fonctions.hpp"

// définir les classe de la std qui seront utilisées
using std::stringstream;
using std::istringstream;
using std::noskipws;
using std::cerr;
using std::endl;

/**
 *  /brief Recherche L'emplacement d'un *char dans un tableau
 *
 *  /param charTab Le tableau de *char dans lequel on fait la recherche
 *  /param taille La taille du tableau
 *  /param cible La chaine de caractère que l'on cherche dans charTab
 *
 * \return L'indice de la cible dans le tableau, si la cible n'est pas présente retourne -1
 */
int index(char** const& charTab, int taille, char const cible[])
{
    for(int i = 0; i < taille; i++)
    {
        if(strcmp(charTab[i],cible) == 0)
        {
            return(i);
        }
    }
    return(-1);
}

/**
 *  /brief Séparer un chaine de caractére en fonction de ';' et '\n' et stocker
 *         le résultat dans un vecteur
 * 
 *  /param str          string à séparer
 *  /param split_str    vecteur dans lequel sauvegarder les résultats
*/

static void lowercase (char* p)
{
	for (; *p != 0; ++p)
		if ('A' <= *p && *p <= 'Z')
			*p += 'a' - 'A';
}

void split_param(const char* str, vector<string>& split_str)
{
    char buffer[1024];
    char* p = buffer;

    const char* c = str;

    for(;;)
    {
        switch(*c)
        {
            // stopper la fonction en arrivant à la fin du string
            case '\0':
                *p = '\0';                
                if(buffer[0] != '\0') {
					lowercase (buffer);
                    split_str.push_back(buffer);
				}
                return;

            // split en fonction du '\n' et ';'
            case '\n':
            case ';':
                *p = '\0';
                if(buffer[0] != '\0') {
					lowercase (buffer);
                    split_str.push_back(buffer);
				}
                p = buffer;
                break;

            // prendre ne compte les escapes
            case '\\':
                ++c;
            default:
                (*p++) = *c;
        }
        ++c;
    }
}

/**
 *  /brief Séparer un chaine de caractére en fonction de '+' à condition
 *         qu'il y ait au moins un espace avant et après le '+'
 * 
 *  /param str          string à séparer
 *  /param split_str    vecteur dans lequel sauvegarder les résultats
*/
void split_metabolites(const char* str, vector<string>& split_str)
{
    // le séparateur est '+'
    // Il faut au moins un espace avant et un après pour qu'il soit considéré
    // comme un séparateur

    char buffer[1024];
    char* p = buffer;

    // c va parcourir les caractères de str
    const char* c = str;

    // un compteur d'espace
    int espace = 0;

    for(;;)
    {
        switch(*c)
        {
            // stopper la fonction en arrivant à la fin du string
            case '\0':
                *p = '\0';
                // ne pas ajouter de buffer vide
                if(buffer[0] != '\0')
                    split_str.push_back(buffer);
                return;

            // split en fonction du '+'
            case '+':
                // s'il n'y a pas d'espace avant ou aprés le '+'
                if(espace == 0 || *(c+1) != ' ')
                {
                    (*p++) = *c;
                }
                else
                {
                    *p = '\0';
                    split_str.push_back(buffer);
                    p = buffer;
                    espace = 0;
                }
                break;

            // prendre en compte les espaces
            case ' ':
                espace++;
                break;

            // prendre en compte les escapes
            case '\\':
                ++c;
            default:
                // ne pas ajouter d'espace au début de la chaine de caractère
                if(p > buffer)
                {
                    while(espace)
                    {
                        (*p++) = ' ';
                        --espace;
                    }
                }
                else
                {
                    espace = 0;
                }
                (*p++) = *c;
        }
        ++c;
    }
}

/**
 *  /brief Segmentation d'une chaine de caractere en un vecteur de string en fonction d'un delimiteur
 *
 *  /param str string a segmenter
 *  /param delimiteur caractère pour segmenter str
 *
 *  /return split_str vecteur de string
 */
vector<string> split(string const& str, char delimiteur)
{
    vector<string> split_str;
    string match;
    stringstream ss(str);

    // Lecture du texte ligne par ligne en separent les lignes par le delimiteur
    while(getline(ss, match, delimiteur))
    {
        // Stocke la "ligne" dans split_str
        split_str.push_back(match);
    }
    return(split_str);
}

/**
 *  /brief Segmentation d'une chaine de caractere en un vecteur de string en fonction d'un delimiteur
 *
 *  /param split_str vecteur qui va stocker les éléments spliter
 *  /param str string a segmenter
 *  /param delimiteur string pour segmenter str
 */
vector<string> split(string const& str, string delimiteur)
{
    string match;
    vector<string> split_str;
    int indexDebut = 0;
    int indexFin = str.find(delimiteur, indexDebut);

    while( indexFin > 0) // indexFin = -1 si il n'y a plus de delimiteur a chercher
    {
        // stockage de la sous chaine de caractere entre indexDebut et indexFin
        match = str.substr(indexDebut, indexFin - indexDebut);
        split_str.push_back(match);
        // IndexDebut prend la valeur de indexFin PLUS la longueur de delimiter !!!
        indexDebut = indexFin + delimiteur.size();
        // IndexFin prend la valeur de l'indice du premier caractere du delimiteur - ou -1 s'il n'y en a plus
        indexFin = str.find(delimiteur, indexDebut);
    }
    if(indexDebut < int(str.size())) // on stocke se qu'il reste
    {
        match = str.substr(indexDebut);
        split_str.push_back(match);
    }
    return(split_str);
}


/**
 *  /brief test si un string est un float
 *
 *  /param mystring string à tester
 *
 *  /return true si c'est un float sinon false
 */
bool estFloat( string const& myString ) {
    istringstream iss(myString);
    float f;
    iss >> noskipws >> f; // pas d'espace !
    return iss.eof() && !iss.fail();
}

#include <strings.h>

int ci_string_compare (const string& s1, const string& s2)
{
	return strcasecmp (s1.c_str (), s2.c_str ());
}

/**
 *  /brief Permet de vérifier que deux vecteurs contiennent la même information
 *         pas forcement dans le même ordre
 *
 *  /param v1 premier vecteur pour la comparaison
 *  /param v2 second vecteur pour la comparaison
 *
 *  /return true si l'information des deux vecteurs est la même
 */
bool memeContenu(vector<string> & v1, vector<string> v2)
{
    if(v1.size() == v2.size())
    {
        sort(v1.begin(), v1.end());
        sort(v2.begin(), v2.end());
        for( size_t i = 0; i < v1.size(); i++)
        {
			if (0 != ci_string_compare (v1[i], v2[i]))
                return false;
        }
        return true;
    }
    else
        return false;
}

/**
 *  /brief permet de savoir si deux vecteurs ont au moins 1 element en commun
 *
 *  /param v1 premier vecteur pour la comparaison
 *  /param v2 second vecteur pour la comparaison
 *
 *  /return true si au moins element est présent dans les deux vecteurs
 */
bool elementCommun(vector<string> const& v1, vector<string> const& v2)
{
    for(string m1 : v1)
        for(string m2 : v2)
        {
			if (0 == ci_string_compare (m1, m2))
                return true;
        }
    return false;
}

/**
 * \brief Convertion d'un entier en string
 * \param i int à  convertir
 * \return string correspondant à i
 */
string itos(int i)
{
    stringstream ss;
    ss << i;
    return(ss.str());
}


/**
 * \brief cherche si la cible est present dans le vecteur
 * \param cible chaine de caractère a chercher
 *        vecteur dans lequel la recherche est faite
 * \return bool disant si oui ou non la cible est contenue dans le vecteur
 */
bool estPresent(string const& cible, vector<string> const& vecteur)
{
    // Parcours le vecteur jusqu'a trouver la cible
    for(size_t i = 0; i < vecteur.size(); i++)
		if (0 == ci_string_compare (cible, vecteur[i]))
            return(true);
    return(false); // sinon retourne false
}

/*
    Verification que le caractére donné est une valeru numérique
*/
bool estNumerique(char const& c)
{
    switch(c){
        case '0':
        case '1':
        case '2':
        case '3':
        case '4':
        case '5':
        case '6':
        case '7':
        case '8':
        case '9':
            return true;
        default:
            return false;
    }
        
}

/*
    Permet de stocker les inhibiteur specifique ou non d'une reaction
 
    ec      code ec de l'enzyme catalusant la reaction 
    org     organisme ou la réaction se déroule in vivo
    vecteurInhibiteur   liste d'inhibiteurs trouvés
    vi_S    liste à remplir d'inhibiteurs inhibant l'enzyme dans 
            l'organisme pour l'EC et l'organisme donné
    vi_S    liste à remplir d'inhibiteurs inhibant uniquement l'enzyme
*/
void estInhibee(string const& ec, string const& org,
                vector<Inhibiteur> const& vecteurInhibiteur,
                vector<Inhibiteur>& vi_S,
                vector<Inhibiteur>& vi_NS)
{
    for(Inhibiteur const& inhi : vecteurInhibiteur)
    {
        if(ec == inhi._ec)
        {
            if(org == inhi._organisme)
            {
                vi_S.push_back(inhi);
            }
            else // org != inhi._organisme
            {
                vi_NS.push_back(inhi);
            }
        }
    }
}


/*
    Insertion d'une réaction dans un vecteur si elle n'est pas deja
    présente
 
    vReac       vecteur qui stocke des réactions
    reaction    reaction à stocker
 */
int insertReaction(vector<Reaction> & vReac, Reaction & reaction)
{
    if( vReac.size() == 0)
        vReac.push_back(reaction);
    else
    {
        vector<Reaction>::iterator it = vReac.begin();
        while(it != vReac.end() && *it < reaction)
        {
            it++;
        }
        if( it != vReac.begin() )
        {
            if(*prev(it) != reaction)
                vReac.insert(it, reaction);
            else
                return 0;
        }
        else
        {
            vReac.insert(it, reaction);
        }
    }
    return 1;
}
/*
    Cette version de la fonction vReac est un vecteur de pointeur de réaction
*/
int insertReaction(vector<Reaction*> & vReac, Reaction & reaction)
{
    if( vReac.size() == 0)
        vReac.push_back(&reaction);
    else
    {
        vector<Reaction*>::iterator it_r;
        it_r = find(vReac.begin(), vReac.end(), &reaction);
        if(it_r == vReac.end())
        {
            vector<Reaction*>::iterator it = vReac.begin();
            while(it != vReac.end() && **it < reaction)
            {
                it++;
            }
            if( it != vReac.begin() )
            {
                if(**prev(it) != reaction)
                    vReac.insert(it, &reaction);
                else
                    return 0;
            }
            else
            {
                vReac.insert(it, &reaction);
            }
        }
    }
    return 1;
}

/*
    Dans cette un ensemble de réactions (reactions) est inséré dans vReac
*/
int insertReaction(vector<Reaction*> & vReac, vector<Reaction*> & reactions)
{
    int nbInsert = 0;
    for(Reaction* reac : reactions)
    {       
        if( vReac.size() == 0)
        {
            vReac.push_back(reac);
            nbInsert += 1;
        }
        else
        {
            vector<Reaction*>::iterator it_r;
            it_r = find(vReac.begin(), vReac.end(), reac);
            if(it_r == vReac.end())
            {
                vector<Reaction*>::iterator it = vReac.begin();
                while(it != vReac.end() && **it < *reac)
                {
                    it++;
                }
                if( it != vReac.begin() )
                {
                    if(**prev(it) != *reac)
                    {
                        vReac.insert(it, reac);
                        nbInsert += 1;
                    }
                    else
                        nbInsert += 0;
                }
                else
                {
                    vReac.insert(it, reac);
                    nbInsert += 1;
                }
            }
        }
    }
    return nbInsert;
}

/*
    permet d'ajouter un métabolite dans un vecteur de métabolite

    vMeta   vecteur de métabolites
    metabolite  le métabolite à ajouter
*/
void ajoutMetabolite(vector<Metabolite> & vMeta, Metabolite const& metabolite)
{
    bool insertion = false;
    for(Metabolite& m : vMeta)
    {
        if(m == metabolite)
        {
            m._reactions_amonts.insert(metabolite._reactions_amonts.begin(),
                                      metabolite._reactions_amonts.end());
            insertion = true;
            break;
        }
    }
    if(!insertion)
    {
        vMeta.push_back(metabolite);
    }
}


void ecritureListeReaction(vector<Reaction*> const& reactionsFinales,
                           string& nomListe)
{
    ofstream liste(nomListe);
    if(liste)
    {
        for(auto reaction : reactionsFinales)
        {
            // Colonne 1 - code EC
            liste << reaction->_ec << "\t";

            // Colonne 2 - Nom de l'enzyme
            liste << reaction->_nomEnzyme << "\t";

            // Colonne 3 - Organisme
            liste << reaction->_organisme << "\t";
            
            // Colonne 4 - Substrats
            ajout_vecteur_flux(reaction->_substrats, liste, "-+-");
            liste << "\t";
            
            // Colonne 5 - Produits
            ajout_vecteur_flux(reaction->_produits, liste, "-+-");
            liste << "\t" ;

            // Colonne 6 - Reversibilité
            if(reaction->_reversible)
                liste << "r" << "\t";
            else
                liste << "?" << "\t";       
                    
            // Colonne 7 - KM
            ajout_vecteur_flux(reaction->_km, liste, ";");
            liste << "\t";
                    
            // Colonne 8 - Kcat
            ajout_vecteur_flux(reaction->_kcat, liste, ";");
            liste << "\t";
            
            // Colonne 9 - pH
            ajout_vecteur_flux(reaction->_ph, liste, "-");
            
            liste << "\n";
        }
        liste.close();
    }
    else
    {
        cerr << "Unable to write to file " << nomListe << endl;
    }
}




void ajout_vecteur_flux(vector<string> const& vect, ofstream& flux, 
                        string sep)
{
    bool premier = true;
    for(auto val : vect)
    {
        if(premier)
        {
            premier = false;
            flux << val;
        }
        else
        {
            flux << sep << val;
        }
    }
}

void ajout_vecteur_flux(vector<float> const& vect, ofstream& flux, 
                        string sep)
{
    bool premier = true;
    for(auto val : vect)
    {
        if(premier)
        {
            premier = false;
            if(val == CONSTANTE_INCONNUE)
                flux << "?";
            else
                flux << val;
        }
        else
        {
            if(val == CONSTANTE_INCONNUE)
                flux << sep << "?";
            else
                flux << sep << val;         
        }
    }
}

/*
Afficher le nombre de metabolites à chaque niveau
*/

void comptage_metabolites_intermediaires(Recherche & recherche)
{

    int c_re1 = 0;
    int c_re2 = 0;
    int c_rs1 = 0;
    int c_rs2 = 0;

    int profondeur = recherche._requete->_profondeur;

    // identifier les metabolites intermediaires des reactions d'entrées
    set<string> produits_entrees_n1;
    set<string> produits_entrees_n2;

    int c_pe1 = 0;
    int c_pe2 = 0;
    for(Reaction& reactionE : recherche._reactionsEntrees[0])
    {
        c_re1++;
        for( string produit : reactionE._produits)
        {
            // H2O est toujours en excé dans le milieu de réaction donc il ne faut pas le prendre en compte
            if(produit != "H2O") 
            {
                produits_entrees_n1.insert(produit);
                c_pe1++;
            }
        }
    }

    cerr << "Number of different products of level 1 input reactions = " << produits_entrees_n1.size() << endl;
    cerr << "Total number of products of the level 1 input reactions = " << c_pe1 << endl;
    cerr << endl;

    if(profondeur == 4)
    {
        for(Reaction& reactionE : recherche._reactionsEntrees[1])
        {
            c_re2++;
            for( string produit : reactionE._produits)
            {
                // H2O est toujours en excé dans le milieu de réaction donc il ne faut pas le prendre en compte
                if(produit != "H2O")
                {
                    produits_entrees_n2.insert(produit);
                    c_pe2++;
                }
            }
        }

        cerr << "Number of different products of the level 2 input reactions = " << produits_entrees_n2.size() << endl;
        cerr << "Total number of products of the level 2 input reactions = " << c_pe2 << endl;
        cerr << endl;
    }

    // for(auto& prod : produits_entrees_n2)
    //     std::cout << prod << endl;

    //identifier les metabolites intermediaires des reactions de sorties
    set<string> substrats_sorties_n1;
    set<string> substrats_sorties_n2;
    set<string> inhibiteurs_sorties_n1;
    set<string> inhibiteurs_sorties_n2;

    int c_ss1 = 0;
    int c_ss2 = 0;
    int c_is1 = 0;
    int c_is2 = 0;

    if(profondeur > 1)
    {
        for(Reaction& reactionS : recherche._reactionsSorties[0])
        {
            c_rs1++;
            // les substrats
            for( string substrat : reactionS._substrats)
            {
                // H2O est toujours en excé dans le milieu de réaction donc il ne faut pas le prendre en compte
                if(substrat != "H2O")
                {
                    substrats_sorties_n1.insert(substrat);
                    c_ss1++;
                }
            }

            //les inhibiteurs
            for(Inhibiteur& inhibiteur : reactionS._inhibiteurs_NS)
            {
                if(inhibiteur._nom != "H2O")
                {
                    inhibiteurs_sorties_n1.insert(inhibiteur._nom);
                    c_is1++;
                }
            }
            for(Inhibiteur& inhibiteur : reactionS._inhibiteurs_S)
            {
                if(inhibiteur._nom != "H2O")
                {
                    inhibiteurs_sorties_n1.insert(inhibiteur._nom);
                    c_is1++;
                }
            }
        }

        cerr << "Number of different substrates of level 1 output reactions = " << substrats_sorties_n1.size() << endl;
        cerr << "Total number of substrates of level 1 output reactions = " << c_ss1 << endl;
        cerr << "Number of different inhibitors of level 1 output reactions = " << inhibiteurs_sorties_n1.size() << endl;
        cerr << "Total number of inhibitors of level 1 output reactions = " << c_is1 << endl;

        substrats_sorties_n1.insert(inhibiteurs_sorties_n1.begin(), inhibiteurs_sorties_n1.end());
        // cerr << "Nombre metabolites des reactions de sortie de niveau 1 = " << substrats_sorties_n1.size() << endl;

        cerr << endl;
    }
    if(profondeur > 2)
    {
        for(Reaction& reactionS : recherche._reactionsSorties[1])
        {
            c_rs2++;
            // les substrats
            for( string substrat : reactionS._substrats)
            {
                // H2O est toujours en excé dans le milieu de réaction donc il ne faut pas le prendre en compte
                if(substrat != "H2O")
                {
                    substrats_sorties_n2.insert(substrat);
                    c_ss2++;
                }
            }


            //les inhibiteurs
            for(Inhibiteur& inhibiteur : reactionS._inhibiteurs_NS)
            {
                if(inhibiteur._nom != "H2O")
                {
                    inhibiteurs_sorties_n2.insert(inhibiteur._nom);
                    c_is1++;
                }
            }
            for(Inhibiteur& inhibiteur : reactionS._inhibiteurs_S)
            {
                if(inhibiteur._nom != "H2O")
                {
                    inhibiteurs_sorties_n2.insert(inhibiteur._nom);
                    c_is2++;
                }
            }
        }

        cerr << "Number of different substrates of level 2 output reactions = " << substrats_sorties_n2.size() << endl;
        cerr << "Total number of substrates of level 2 output reactions = " << c_ss2 << endl;
        cerr << "Number of different inhibitors of level 1 output reactions = " << inhibiteurs_sorties_n2.size() << endl;
        cerr << "Total number of inhibitors of level 1 output reactions = " << c_is1 << endl;

        substrats_sorties_n2.insert(inhibiteurs_sorties_n2.begin(), inhibiteurs_sorties_n2.end());
        // cerr << "Nombre metabolites des reactions de sortie de niveau 2 = " << substrats_sorties_n2.size() << endl;
        cerr << endl;
    }


    cerr << "Number of level 1 input reactions =" << c_re1 << endl;
    cerr << "Number of level 2 input reactions =" << c_re2 << endl;
    cerr << "Number of level 1 output reactions =" << c_rs1 << endl;
    cerr << "Number of level 2 output reactions =" << c_rs2 << endl;
}