#include <iostream>
#include <fstream>
#include <cstring> // strcmp()

#include "Requete.hpp"
#include "fonctions.hpp" // split()
#include "Recherche.hpp"
#include "Traitement.hpp"
#include "ecritureReseau.hpp"

#include <chrono>


std::chrono::steady_clock::time_point end = std::chrono::steady_clock::now();

using std::cerr;
using std::endl;
using std::stoi; // <string>

bool	use_ec_number = false;		// wether use EC number rather than enzyme name.

int main(int argc, char** argv) 
{
    // objet pour la gestion des paramètres
    Requete requete = Requete();
    if(argc > 1)
    {
        // --------------------------------------------------------------------
        // Gestion des paramètres
        // --------------------------------------------------------------------
        if(Requete::_verbose)
            cerr << "-> Parameter management" << endl;
        
        // l'utilisateur veut il afficher l'aide
        if(index(argv, argc, "-h") != -1 || index(argv, argc, "--help") != -1)
            requete.afficher_aide(argv[0]); // le programme s’arrêtera

        // parcours des arguments à générer
        for(int i = 1; i < argc; ++i)
        {
            // Les entrées du réseau
            if(strcmp(argv[i], "-in")        == 0)
            {
                if(i == argc -1)
                {
                    cerr << "- missing value for -in\n";
                    exit(0);
                }
                else
                    split_param(argv[++i], requete._entrees);
            }

            // les sorties du réseau à générer
            else if(strcmp(argv[i], "-out")  == 0)                
            {
                if(i == argc -1)
                {
                    cerr << "- missing value for -out\n";
                    exit(0);
                }
                else
                    split_param(argv[++i], requete._sorties);
            }

            // La profondeur du réseau à générer (nombre de réaction entre les entrées et sorties)
            else if(strcmp(argv[i], "-d")    == 0)
            {
                if(i == argc -1)
                {
                    cerr << "- missing value for -p\n";
                    exit(0);
                }
                else
                    requete.recupererProfondeur(argv[++i]);
            }

            // Est-ce qu'il faut faire la recherche dans les niveaux inférieure
            else if(strcmp(argv[i], "-sd")   == 0 || strcmp(argv[i], "--smaller-distance")   == 0)
                requete._rechercheProfondeursInferieures = true;

            // Nom de la banque de réactions dans laquelle chercher
            else if(strcmp(argv[i], "-br")   == 0)
            {
                if(i == argc -1)
                {
                    cerr << "- missing value for -br\n";
                    exit(0);
                }
                else
                    requete._nom_banqueReactions = argv[++i];
            }

            // Nom de la banque d'inhibition
            else if(strcmp(argv[i], "-bi")   == 0)
            {
                if(i == argc -1)
                {
                    cerr << "- missing value for -bi\n";
                    exit(0);
                }
                else
                {
                    requete._nom_banqueInhibitions = argv[++i];
                    requete._rechercheInhibitions  = true;
                }
            }

            // Faire un filtre sur la valeur des km
            else if(strcmp(argv[i], "-km")   == 0)
            {
                if(i == argc -1)
                {
                    cerr << "- missing value for -km\n";
                    exit(0);
                }
                else
                    requete.intervalleConstante(KM, argv[++i]);
            }

            // faire un filtre sur la valeur des kcat
            else if(strcmp(argv[i], "-kcat") == 0)
            {
                if(i == argc -1)
                {
                    cerr << "- missing value for -kcat\n";
                    exit(0);
                }
                else
                    requete.intervalleConstante(KCAT, argv[++i]);
            }

            // faire un filtre sur le ph
            else if(strcmp(argv[i], "-ph")   == 0)
            {
                if(i == argc -1)
                {
                    cerr << "- missing value for -ph\n";
                    exit(0);
                }
                else
                    requete.intervalleConstante(PH, argv[++i]);
            }

            // faire un filtre sur le ki
            else if(strcmp(argv[i], "-ki")   == 0)
            {
                if(i == argc -1)
                {
                    cerr << "- missing value for -ki\n";
                    exit(0);
                }
                else
                    requete.intervalleConstante(KI, argv[++i]);
            }

            // faire un filtre sur la valeur d'IC50
            else if(strcmp(argv[i], "-ic50") == 0)
            {
                if(i == argc -1)
                {
                    cerr << "- missing value for -ic50\n";
                    exit(0);
                }
                else
                    requete.intervalleConstante(IC50, argv[++i]);
            }

            // nom du fichier qui contiendra le réseau de réactions
            else if(strcmp(argv[i], "-o")    == 0)
            {
                if(i == argc -1)
                {
                    cerr << "- missing value for -o\n";
                    exit(0);
                }
                else
                    requete._fichierSortie = argv[++i];
            }

            // Le réseau doit être généré en MTO
            else if(strcmp(argv[i], "-mto")  == 0)
                requete._metatool = true;

            // Nom du fichier qui contiendra la liste des réactions avec leurs 
            else if(strcmp(argv[i], "-l")    == 0)
            {
                if(i == argc -1)
                {
                    cerr << "- missing value for -l\n";
                    exit(0);
                }
                else
                    requete._listeSortie = argv[++i];
            }

            // afficher des informations complémentaire durant l'execution
            else if(strcmp(argv[i], "-v") == 0 || strcmp(argv[i], "--verbose") == 0)
                requete._verbose = true;


            // Par défaut écrire un SBML valide d'après sbml.org
            // avec cette option ecrite un sbml compatible avec le programme scmaker
            //
            // Dans <listOfModifiers>
            // modifierSpeciesReferences -> speciesReference
            else if(strcmp(argv[i], "-wmt") == 0)
                requete._wrong_modifier_tag = true;

            // donner le temps de recherche de reaction et de connexion
            else if(strcmp(argv[i], "-chrono") == 0)
                requete._chrono = true;

            // use EC numbers rather than enzymes name.
            else if(strcmp(argv[i], "-ec") == 0)
                use_ec_number = true;

            // donner le temps de recherche de reaction et de connexion
            else if(strcmp(argv[i], "-stat") == 0)
                requete._stat = true;

            else
            {
                cerr << "error: Parameter: " << argv[i] << " does not exist\n"
                     << "To display help, use the -h parameter\n";
                exit(0);
            }
        }

        // commande pour vérifier que les valeurs des paramètres sont correctes
        // dans le cas contraire afficher les problèmes
        if(!requete.commandeValide())
            exit(0);

        // Si tout va bien alors on peut charger les fichiers dans la mémoires
        requete.decompression("reactions");
        if(requete._rechercheInhibitions)
            requete.decompression("inhibitions");


        // --------------------------------------------------------------------
        // Recherche de réactions et inhibitions
        // --------------------------------------------------------------------
        std::chrono::steady_clock::time_point begin = std::chrono::steady_clock::now();


        if(Requete::_verbose)
            cerr << "-> Searching for reactions and inhibitions" << endl;
        Recherche recherche = Recherche(requete);
        Traitement traitement;
        if(Requete::_stat)
        {
            if(Requete::_chrono && Requete::_verbose)
            {
                std::chrono::steady_clock::time_point end = std::chrono::steady_clock::now();
                cerr << "Time (ms) = " << std::chrono::duration_cast<std::chrono::milliseconds>(end-begin).count() << endl;
            }
            comptage_metabolites_intermediaires(recherche);
        }
        else
        {
            if(requete._rechercheSortie)
            {
                // Traitement des réactions trouvées
                if(Requete::_verbose)
                    cerr << "    -> Processing of reactions found" << endl;
                traitement = Traitement(recherche);
            }
            else
            {
                for(vector<Reaction>& vr : recherche._reactionsEntrees)
                {
                    for(Reaction& r : vr)
                    {
                        traitement._reactionsFinales.push_back(&r);
                    }
                }
            }

            if(Requete::_verbose)
            {
                std::chrono::steady_clock::time_point end = std::chrono::steady_clock::now();
                cerr << "\n\nTime (ms) = " << std::chrono::duration_cast<std::chrono::milliseconds>(end-begin).count() << endl;
            }
        }
        // --------------------------------------------------------------------
        // Ecriture de la liste de réactions sélectionnées
        // --------------------------------------------------------------------
        if(requete._fichierSortie != "NULL")
        {
            if(requete._listeSortie.size() > 0)
                ecritureListeReaction(traitement._reactionsFinales, requete._listeSortie);

            // Ecriture du/des fichier(s) de sortie 
            if(requete._metatool)
                ecritureReseau_mto(traitement._reactionsFinales, requete._fichierSortie);   
            else
                ecritureReseau_xml(traitement._reactionsFinales, requete._fichierSortie);
        }
    }
    else
    {
        requete.afficher_aide(argv[0]); // le programme s'arrêtera
    }
}